@@ -43,6 +43,7 @@ import os
 import smtplib
 import sys
 import traceback
+import functools
 from datetime import datetime
 from email.mime.multipart import MIMEMultipart
 from email.mime.text import MIMEText
@@ -50,7 +51,7 @@ from email.mime.image import MIMEImage
 from io import open
 from importlib import import_module
 from multiprocessing import Pool
-from socket import gethostname, gethostbyname
+from socket import gethostname, gethostbyname, getaddrinfo, getfqdn
 from smtplib import (SMTPRecipientsRefused, SMTPHeloError, SMTPSenderRefused,
                      SMTPDataError)
 
@@ -61,10 +62,16 @@ from gridmap.conf import (CHECK_FREQUENCY, CREATE_PLOTS, DEFAULT_QUEUE,
                           ERROR_MAIL_SENDER, HEARTBEAT_FREQUENCY,
                           IDLE_THRESHOLD, MAX_IDLE_HEARTBEATS,
                           MAX_TIME_BETWEEN_HEARTBEATS, NUM_RESUBMITS,
-                          SEND_ERROR_MAIL, SMTP_SERVER, USE_MEM_FREE)
+                          SEND_ERROR_MAIL, SMTP_SERVER, USE_MEM_FREE,
+                          DEFAULT_TEMP_DIR, DEFAULT_PAR_ENV)
 from gridmap.data import zdumps, zloads
 from gridmap.runner import _heart_beat
 
+
+class DRMAANotPresentException(ImportError):
+    pass
+
+
 if DRMAA_PRESENT:
     from drmaa import (ExitTimeoutException, InvalidJobException,
                        JobControlAction, JOB_IDS_SESSION_ALL, Session,
@@ -103,15 +110,22 @@ class Job(object):
        are defined at the module or class level).
     """
 
-    __slots__ = ('_f', 'args', 'id', 'kwlist', 'cleanup', 'ret', 'traceback',
+    __slots__ = ('function', 'args', 'id', 'kwlist', 'cleanup', 'ret', 'traceback',
                  'num_slots', 'mem_free', 'white_list', 'path', 'uniq_id',
                  'name', 'queue', 'environment', 'working_dir',
                  'cause_of_death', 'num_resubmits', 'home_address',
                  'log_stderr_fn', 'log_stdout_fn', 'timestamp', 'host_name',
-                 'heart_beat', 'track_mem', 'track_cpu')
+                 'heart_beat', 'track_mem', 'track_cpu', 'interpreting_shell',
+                 'copy_env', 'add_env', 'project', 'validation_level', 
+                 'os_distribution', 'os_minor', 'par_env', 'gpu', 'h_vmem', 'h_rt',
+                 'resources')
 
     def __init__(self, f, args, kwlist=None, cleanup=True, mem_free="1G",
-                 name='gridmap_job', num_slots=1, queue=DEFAULT_QUEUE):
+                 name='gridmap_job', num_slots=1, queue=DEFAULT_QUEUE,
+                 interpreting_shell=None, copy_env=True, add_env=None,
+                 project=None, validation_level=None, os_distribution=None,
+                 os_minor=None, par_env=DEFAULT_PAR_ENV, gpu=0, h_vmem=None,
+                 h_rt=None, resources=None):
         """
         Initializes a new Job.
 
@@ -132,8 +146,34 @@ class Job(object):
         :type num_slots: int
         :param queue: SGE queue to schedule job on.
         :type queue: str
-
+        :param interpreting_shell: The interpreting shell for the job
+        :type interpreting_shell: str
+        :param copy_env: copy environment from master node to worker node?
+        :type copy_env: boolean
+        :param add_env: Environment variables to add to the environment.
+                        Overwrites variables which already exist due to
+                        ``copy_env=True``.
+        :type add_env: dict
+        :param par_env: parallel environment to use.
+        :type par_env: str
+        :param project: the project to which this job is assigned
+        :type project: str
+        :param validation_level: validation level for the job
+        :type validation_level: str value e|w|n|p|v
+        :param os_distribution: os that need job to run on machine
+        :type os_distribution: str
+        :param os_minor: os minor version that need job to run on machine
+        :type os_minor: str
+        :param gpu: number of GPUs to request
+        :type gpu: int
+        :param h_vmem: hard virtual memory limit (e.g. "4G")
+        :type h_vmem: str, optional
+        :param h_rt: hard runtime limit (e.g. "00:59:00")
+        :type h_rt: str, optional
+        :param resources: list of additional custom resources specifications
+        :type resources: list of str, optional
         """
+        avaliable_validation_levels = ['e', 'w', 'n', 'p', 'v'] 
         self.track_mem = []
         self.track_cpu = []
         self.heart_beat = None
@@ -146,7 +186,6 @@ class Job(object):
         self.num_resubmits = 0
         self.cause_of_death = ''
         self.path = None
-        self._f = None
         self.function = f
         self.args = args
         self.id = -1
@@ -158,57 +197,38 @@ class Job(object):
         self.white_list = []
         self.name = name.replace(' ', '_')
         self.queue = queue
+        self.interpreting_shell = interpreting_shell
+        self.copy_env = copy_env
         # Save copy of environment variables
         self.environment = {}
-        for env_var, value in os.environ.items():
-            try:
-                if not isinstance(env_var, bytes):
-                    env_var = env_var.encode()
-                if not isinstance(value, bytes):
-                    value = value.encode()
-            except UnicodeEncodeError:
-                logger = logging.getLogger(__name__)
-                logger.warning('Skipping non-ASCII environment variable.')
-            else:
-                self.environment[env_var] = value
+        def _add_env(env_vars):
+            for env_var, value in env_vars.items():
+                try:
+                    if not isinstance(env_var, bytes):
+                        env_var = env_var.encode()
+                    if not isinstance(value, bytes):
+                        value = value.encode()
+                except UnicodeEncodeError:
+                    logger = logging.getLogger(__name__)
+                    logger.warning('Skipping non-ASCII environment variable.')
+                else:
+                    self.environment[env_var] = value
+        if self.copy_env:
+            _add_env(os.environ)
+        if add_env is not None:
+            _add_env(add_env)
         self.working_dir = os.getcwd()
-
-    @property
-    def function(self):
-        ''' Function this job will execute. '''
-        return self._f
-
-    @function.setter
-    def function(self, f):
-        """
-        setter for function that carefully takes care of
-        namespace, avoiding __main__ as a module
-        """
-
-        m = inspect.getmodule(f)
-        try:
-            self.path = os.path.dirname(os.path.abspath(
-                inspect.getsourcefile(f)))
-        except TypeError:
-            self.path = ''
-
-        # if module is not __main__, all is good
-        if m.__name__ != "__main__":
-            self._f = f
-
-        else:
-
-            # determine real module name
-            mn = os.path.splitext(os.path.basename(m.__file__))[0]
-
-            # make sure module is present
-            import_module(mn)
-
-            # get module
-            mod = sys.modules[mn]
-
-            # set function from module
-            self._f = getattr(mod, f.__name__)
+        self.par_env = par_env
+        self.project = project
+        self.validation_level = validation_level
+        if self.validation_level and not self.validation_level in avaliable_validation_levels:
+            raise ValueError("Validation level can be only e|w|n|p|v")
+        self.os_distribution = os_distribution
+        self.os_minor = os_minor
+        self.gpu = gpu
+        self.h_vmem = h_vmem
+        self.h_rt = h_rt
+        self.resources = resources
 
     def execute(self):
         """
@@ -231,16 +251,35 @@ class Job(object):
         define python-style getter
         """
 
-        ret = "-shell yes -b yes"
+        ret = "-shell yes"
+        if self.interpreting_shell:
+            ret += " -S {}".format(self.interpreting_shell)
+        ret += " -b yes"
 
         if self.mem_free and USE_MEM_FREE:
             ret += " -l mem_free={}".format(self.mem_free)
         if self.num_slots and self.num_slots > 1:
-            ret += " -pe smp {}".format(self.num_slots)
+            ret += " -pe {} {}".format(self.par_env, self.num_slots)
         if self.white_list:
             ret += " -l h={}".format('|'.join(self.white_list))
         if self.queue:
             ret += " -q {}".format(self.queue)
+        if self.project:
+            ret += " -P {}".format(self.project)
+        if self.validation_level:
+            ret += " -w {}".format(self.validation_level)
+        if self.os_distribution:
+            ret += " -l os_distribution={}".format(self.os_distribution)
+        if self.os_minor:
+            ret += " -l os_minor={}".format(self.os_minor)
+        if self.gpu:
+            ret += " -l gpu={}".format(self.gpu)
+        if self.h_vmem:
+            ret += " -l h_vmem={}".format(self.h_vmem)
+        if self.h_rt:
+            ret += " -l h_rt={}".format(self.h_rt)
+        if self.resources:
+            ret += "".join([" -l {}".format(x) for x in self.resources])
 
         return ret
 
@@ -253,7 +292,7 @@ class JobMonitor(object):
     """
     Job monitor that communicates with other nodes via 0MQ.
     """
-    def __init__(self, temp_dir='/scratch'):
+    def __init__(self, temp_dir=DEFAULT_TEMP_DIR):
         """
         set up socket
         """
@@ -265,6 +304,18 @@ class JobMonitor(object):
 
         self.host_name = gethostname()
         self.ip_address = gethostbyname(self.host_name)
+
+        for _, _, _, _, (ip, _) in getaddrinfo(getfqdn(), 0):
+            if ip != '127.0.0.1':
+                self.ip_address = ip
+                self.interface = "tcp://%s" % (self.ip_address)
+                break
+        else:
+            self.logger.warning('IP address for JobMonitor server is '
+                                '127.0.0.1.  Runners on other machines will be'
+                                ' unable to connect.')
+            self.ip_address = '127.0.0.1'
+
         self.interface = "tcp://%s" % (self.ip_address)
 
         # bind to random port and remember it
@@ -508,26 +559,58 @@ class JobMonitor(object):
                    for job in self.jobs)
 
 
-def send_error_mail(job):
+def _send_mail(subject, body_text, attachments=None):
     """
-    send out diagnostic email
+    Send out job status email
+
+    This is a helper function for send_error_mail and send_completion_mail
     """
+
     logger = logging.getLogger(__name__)
 
     # Connect to server
     try:
         s = smtplib.SMTP(SMTP_SERVER)
     except smtplib.SMTPConnectError:
-        logger.error('Failed to connect to SMTP server to send error ' +
+        logger.error('Failed to connect to SMTP server to send status ' +
                      'email.', exc_info=True)
         return
 
     # create message
     msg = MIMEMultipart()
-    msg["subject"] = "GridMap error {}".format(job.name)
+    msg["subject"] = subject
     msg["From"] = ERROR_MAIL_SENDER
     msg["To"] = ERROR_MAIL_RECIPIENT
 
+    logger.info('Email body: %s', body_text)
+
+    body_msg = MIMEText(body_text)
+    msg.attach(body_msg)
+
+    if attachments is not None:
+        for attachment in attachments:
+            msg.attach(attachment)
+
+    # Send mail
+    try:
+        s.sendmail(ERROR_MAIL_SENDER, ERROR_MAIL_RECIPIENT, msg.as_string())
+    except (SMTPRecipientsRefused, SMTPHeloError, SMTPSenderRefused,
+            SMTPDataError):
+        logger.error('Failed to send status email.', exc_info=True)
+
+    s.quit()
+
+
+def send_error_mail(job):
+    """
+    send out diagnostic email
+    """
+    logger = logging.getLogger(__name__)
+
+    # create message
+    subject = "GridMap error {}".format(job.name)
+    attachments = list()
+
     # compose error message
     body_text = ""
     body_text += "Job {}\n".format(job.name)
@@ -547,17 +630,12 @@ def send_error_mail(job):
         body_text += "Job encountered exception: {}\n".format(job.ret)
         body_text += "Stacktrace: {}\n\n".format(job.traceback)
 
-    logger.info('Email body: %s', body_text)
-
-    body_msg = MIMEText(body_text)
-    msg.attach(body_msg)
-
     # attach log file
     if job.heart_beat and "log_file" in job.heart_beat:
         log_file_attachement = MIMEText(job.heart_beat['log_file'])
         log_file_attachement.add_header('Content-Disposition', 'attachment',
                                         filename='{}_log.txt'.format(job.id))
-        msg.attach(log_file_attachement)
+        attachments.append(log_file_attachement)
 
     # if matplotlib is installed
     if CREATE_PLOTS:
@@ -584,7 +662,7 @@ def send_error_mail(job):
         img_mem_attachement = MIMEImage(img_data)
         img_mem_attachement.add_header('Content-Disposition', 'attachment',
                                        filename=os.path.basename(img_mem_fn))
-        msg.attach(img_mem_attachement)
+        attachments.append(img_mem_attachement)
 
         # attach cpu plot
         img_cpu_fn = os.path.join("/tmp", "{}_cpu.png".format(job.id))
@@ -599,23 +677,18 @@ def send_error_mail(job):
         img_cpu_attachement = MIMEImage(img_data)
         img_cpu_attachement.add_header('Content-Disposition', 'attachment',
                                        filename=os.path.basename(img_cpu_fn))
-        msg.attach(img_cpu_attachement)
+        attachments.append(img_cpu_attachement)
 
     # Send mail
-    try:
-        s.sendmail(ERROR_MAIL_SENDER, ERROR_MAIL_RECIPIENT, msg.as_string())
-    except (SMTPRecipientsRefused, SMTPHeloError, SMTPSenderRefused,
-            SMTPDataError):
-        logger.error('Failed to send error email.', exc_info=True)
+    _send_mail(subject, body_text, attachments)
 
     # Clean up plot temporary files
     if CREATE_PLOTS:
         os.unlink(img_cpu_fn)
         os.unlink(img_mem_fn)
-    s.quit()
 
 
-def handle_resubmit(session_id, job, temp_dir='/scratch/'):
+def handle_resubmit(session_id, job, temp_dir=DEFAULT_TEMP_DIR):
     """
     heuristic to determine if the job should be resubmitted
 
@@ -688,7 +761,7 @@ def _process_jobs_locally(jobs, max_processes=1):
     return jobs
 
 
-def _submit_jobs(jobs, home_address, temp_dir='/scratch', white_list=None,
+def _submit_jobs(jobs, home_address, temp_dir=DEFAULT_TEMP_DIR, white_list=None,
                  quiet=True):
     """
     Method used to send a list of jobs onto the cluster.
@@ -725,7 +798,7 @@ def _submit_jobs(jobs, home_address, temp_dir='/scratch', white_list=None,
     return sid
 
 
-def _append_job_to_session(session, job, temp_dir='/scratch/', quiet=True):
+def _append_job_to_session(session, job, temp_dir=DEFAULT_TEMP_DIR, quiet=True):
     """
     For an active session, append new job based on information stored in job
     object. Also sets job.id to the ID of the job on the grid.
@@ -782,8 +855,8 @@ def _append_job_to_session(session, job, temp_dir='/scratch/', quiet=True):
     session.deleteJobTemplate(jt)
 
 
-def process_jobs(jobs, temp_dir='/scratch/', white_list=None, quiet=True,
-                 max_processes=1, local=False):
+def process_jobs(jobs, temp_dir=DEFAULT_TEMP_DIR, white_list=None, quiet=True,
+                 max_processes=1, local=False, require_cluster=False):
     """
     Take a list of jobs and process them on the cluster.
 
@@ -803,11 +876,18 @@ def process_jobs(jobs, temp_dir='/scratch/', white_list=None, quiet=True,
     :param local: Should we execute the jobs locally in separate processes
                   instead of on the the cluster?
     :type local: bool
+    :param require_cluster: Should we raise an exception if access to cluster
+                            is not available?
+    :type require_cluster: bool
 
     :returns: List of Job results
     """
     if (not local and not DRMAA_PRESENT):
         logger = logging.getLogger(__name__)
+        if require_cluster:
+            raise DRMAANotPresentException(
+                'Could not import drmaa, but cluster access required.'
+            )
         logger.warning('Could not import drmaa. Processing jobs locally.')
         local = True
 
@@ -858,8 +938,12 @@ def _resubmit(session_id, job, temp_dir):
 # MapReduce Interface
 #####################
 def grid_map(f, args_list, cleanup=True, mem_free="1G", name='gridmap_job',
-             num_slots=1, temp_dir='/scratch/', white_list=None,
-             queue=DEFAULT_QUEUE, quiet=True, local=False, max_processes=1):
+             num_slots=1, temp_dir=DEFAULT_TEMP_DIR, white_list=None,
+             queue=DEFAULT_QUEUE, quiet=True, local=False, max_processes=1,
+             interpreting_shell=None, copy_env=True, add_env=None, project=None,
+             validation_level=None, os_distribution=None, os_minor=None, gpu=0,
+             h_vmem=None, h_rt=None, resources=None, completion_mail=False,
+             require_cluster=False, par_env=DEFAULT_PAR_ENV):
     """
     Maps a function onto the cluster.
 
@@ -900,6 +984,38 @@ def grid_map(f, args_list, cleanup=True, mem_free="1G", name='gridmap_job',
     :param max_processes: The maximum number of concurrent processes to use if
                           processing jobs locally.
     :type max_processes: int
+    :param interpreting_shell: The interpreting shell for the jobs.
+    :type interpreting_shell: str
+    :param copy_env: copy environment from master node to worker node?
+    :type copy_env: boolean
+    :param add_env: Environment variables to add to the environment.
+                    Overwrites variables which already exist due to
+                    ``copy_env=True``.
+    :type add_env: dict
+    :param gpu: number of GPUs to request
+    :type gpu: int
+    :param h_vmem: hard virtual memory limit (e.g. "4G")
+    :type h_vmem: str, optional
+    :param h_rt: hard runtime limit (e.g. "00:59:00")
+    :type h_rt: str, optional
+    :param resources: list of additional custom resources specifications
+    :type resources: list of str, optional
+    :param par_env: parallel environment to use.
+    :type par_env: str
+    :param completion_mail: whether to send an e-mail upon completion of all
+                            jobs
+    :type completion_mail: boolean
+    :param require_cluster: Should we raise an exception if access to cluster
+                            is not available?
+    :type require_cluster: bool
+    :param project: project for the job
+    :type project: str
+    :param validation_level: validation level for the job
+    :type validation_level: str value e|w|n|p|v
+    :param os_distribution: os that need job to run on machine
+    :type os_distribution: str
+    :param os_minor: os minor version that need job to run on machine
+    :type os_minor: str
 
     :returns: List of Job results
     """
@@ -908,13 +1024,37 @@ def grid_map(f, args_list, cleanup=True, mem_free="1G", name='gridmap_job',
     jobs = [Job(f, [args] if not isinstance(args, list) else args,
                 cleanup=cleanup, mem_free=mem_free,
                 name='{}{}'.format(name, job_num), num_slots=num_slots,
-                queue=queue)
+                queue=queue, interpreting_shell=interpreting_shell,
+                copy_env=copy_env, add_env=add_env, par_env=par_env,
+                gpu=gpu, h_vmem=h_vmem, h_rt=h_rt, resources=resources,
+                project=project, validation_level=validation_level,
+                os_distribution=os_distribution, os_minor=os_minor)
             for job_num, args in enumerate(args_list)]
 
     # process jobs
-    job_results = process_jobs(jobs, temp_dir=temp_dir, white_list=white_list,
+    job_results = process_jobs(jobs, temp_dir=temp_dir,
+                               white_list=white_list,
                                quiet=quiet, local=local,
-                               max_processes=max_processes)
+                               max_processes=max_processes,
+                               require_cluster=require_cluster)
+
+    # send a completion mail (if requested and configured)
+    if completion_mail and SEND_ERROR_MAIL:
+        send_completion_mail(name=name)
 
     return job_results
 
+
+def send_completion_mail(name):
+    """
+    send out success email
+    """
+    # create message
+    subject = "GridMap completed grid_map {}".format(name)
+
+    # compose error message
+    body_text = ""
+    body_text += "Job {}\n".format(name)
+
+    # Send mail
+    _send_mail(subject, body_text)
