@@ -114,11 +114,11 @@ def get_memory_usage(pid, heart_pid):
     :returns: Total memory usage of process (and children) in Mb.
     """
     process = psutil.Process(pid)
-    mem_total = float(process.get_memory_info()[0])
-    for p in process.get_children(recursive=True):
+    mem_total = float(process.memory_info()[0])
+    for p in process.children(recursive=True):
         if p.is_running() and p.pid != heart_pid:
             try:
-                mem_total += float(p.get_memory_info()[0])
+                mem_total += float(p.memory_info()[0])
             except psutil.NoSuchProcess:
                 continue
     return mem_total / (1024.0 ** 2.0)
@@ -138,14 +138,14 @@ def get_cpu_load(pid, heart_pid):
     :rtype: (float, bool)
     """
     process = psutil.Process(pid)
-    cpu_sum = float(process.get_cpu_percent())
+    cpu_sum = float(process.cpu_percent())
     running = process.status not in _SLEEP_STATUSES
     num_procs = 1
-    for p in process.get_children(recursive=True):
+    for p in process.children(recursive=True):
         # Make sure this process hasn't exited before querying its status
         if p.is_running() and p.pid != heart_pid:
             try:
-                cpu_sum += float(p.get_cpu_percent())
+                cpu_sum += float(p.cpu_percent())
                 running = running or (p.status not in _SLEEP_STATUSES)
             except psutil.NoSuchProcess:
                 continue
@@ -197,7 +197,7 @@ def _run_job(job_id, address):
     heart.start()
 
     wait_sec = random.randint(0, 5)
-    logger.info("Waiting %i seconds before fetching input", wait_sec)
+    logger.info("Waiting %s seconds before fetching input", wait_sec)
     time.sleep(wait_sec)
 
     try:
@@ -272,7 +272,7 @@ def _main():
     logger.info("Appended {0} to PYTHONPATH".format(args.module_dir))
     sys.path.insert(0, args.module_dir)
 
-    logger.debug("Job ID: %i\tHome address: %s\tModule dir: %s",
+    logger.debug("Job ID: %s\tHome address: %s\tModule dir: %s",
                  os.environ['JOB_ID'],
                  args.home_address, args.module_dir)
 
